package ru.dbotthepony.mc.otm.client.render

import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.core.immutableMap
import ru.dbotthepony.mc.otm.core.math.RelativeSide

private fun makeButton(grid: GridAtlas): StretchingRectangleElement {
	val x = grid.currentX
	val y = grid.currentY

	return StretchingRectangleElement(
		topLeft = grid.texture.sprite(x, y, 2f, 2f, grid.atlasWidth, grid.atlasHeight),
		topRight = grid.texture.sprite(x + 16f, y, 2f, 2f, grid.atlasWidth, grid.atlasHeight),
		bottomLeft = grid.texture.sprite(x, y + 16f, 2f, 2f, grid.atlasWidth, grid.atlasHeight),
		bottomRight = grid.texture.sprite(x + 16f, y + 16f, 2f, 2f, grid.atlasWidth, grid.atlasHeight),
		middle = grid.texture.sprite(x + 2f, y + 2f, 14f, 14f, grid.atlasWidth, grid.atlasHeight),
		top = grid.texture.sprite(x + 2f, y, 14f, 2f, grid.atlasWidth, grid.atlasHeight),
		left = grid.texture.sprite(x, y + 2f, 2f, 14f, grid.atlasWidth, grid.atlasHeight),
		right = grid.texture.sprite(x + 16f, y + 2f, 2f, 14f, grid.atlasWidth, grid.atlasHeight),
		bottom = grid.texture.sprite(x + 2f, y + 16f, 14f, 2f, grid.atlasWidth, grid.atlasHeight),
	)
}

object Widgets18 {
	private val buttonGrids = WidgetLocation.LARGE_BUTTON.grid(rows = 1, columns = 4)

	val BUTTON_IDLE_STRETCHABLE = makeButton(buttonGrids)
	val BUTTON_IDLE = buttonGrids.next()
	val BUTTON_HOVERED_STRETCHABLE = makeButton(buttonGrids)
	val BUTTON_HOVERED = buttonGrids.next()
	val BUTTON_PRESSED_STRETCHABLE = makeButton(buttonGrids)
	val BUTTON_PRESSED = buttonGrids.next()
	val BUTTON_DISABLED_STRETCHABLE = makeButton(buttonGrids)
	val BUTTON_DISABLED = buttonGrids.next()

	private val storageGrid = WidgetLocation.STORAGE_CONTROLS.grid(rows = 2, columns = 5)
	val ARROW_DOWN = storageGrid.next()
	val ARROW_UP = storageGrid.next()
	val SORT_DEFAULT = storageGrid.next()
	val SORT_ALPHABET = storageGrid.next()
	val SORT_COUNT = storageGrid.next()
	val SORT_MODID = storageGrid.next()
	val SORT_ID = storageGrid.next()
	val SORT_MATTER_VALUE = storageGrid.next()
	val SORT_MATTER_COMPLEXITY = storageGrid.next()

	private val miscGrid = WidgetLocation.MISC_18.grid(4, 4)

	val COOLDOWN = miscGrid.next()
	val CROSS = miscGrid.next()
	val FORWARD_SLASH = miscGrid.next()
	val RETURN_ARROW_LEFT = miscGrid.next()

	private val slotBgGrid = WidgetLocation.SLOT_BACKGROUNDS.grid(4, 4)

	val BATTERY_SLOT_BACKGROUND = slotBgGrid.next()
	val EQUIPMENT_BATTERY_SLOT_BACKGROUND = slotBgGrid.next()
	val PATTERN_SLOT_BACKGROUND = slotBgGrid.next()
	val MATTER_CAPACITOR_SLOT_BACKGROUND = slotBgGrid.next()

	private val redstoneGrid = WidgetLocation.REDSTONE_CONTROLS.grid(rows = 1, columns = 3)

	val REDSTONE_IGNORED = redstoneGrid.next()
	val REDSTONE_LOW = redstoneGrid.next()
	val REDSTONE_HIGH = redstoneGrid.next()

	private val controlsGrid = WidgetLocation.SIDE_CONTROLS.grid(rows = 4, columns = 8)

	val PULL = controlsGrid.next()
	val PUSH = controlsGrid.next()
	val PULL_DISABLED = controlsGrid.next()
	val PUSH_DISABLED = controlsGrid.next()

	private fun controls() = immutableMap {
		put(FlowDirection.NONE, controlsGrid.next())
		put(FlowDirection.INPUT, controlsGrid.next())
		put(FlowDirection.OUTPUT, controlsGrid.next())
		put(FlowDirection.BI_DIRECTIONAL, controlsGrid.next())
	}

	private fun controls2(input: Map<FlowDirection, MatterySprite>) = immutableMap {
		put(MatteryDeviceBlockEntity.ItemHandlerMode.DISABLED, input[FlowDirection.NONE]!!)
		put(MatteryDeviceBlockEntity.ItemHandlerMode.INPUT, input[FlowDirection.INPUT]!!)
		put(MatteryDeviceBlockEntity.ItemHandlerMode.OUTPUT, input[FlowDirection.OUTPUT]!!)
		put(MatteryDeviceBlockEntity.ItemHandlerMode.INPUT_OUTPUT, input[FlowDirection.BI_DIRECTIONAL]!!)
		put(MatteryDeviceBlockEntity.ItemHandlerMode.BATTERY, BATTERY_ONLY)
	}

	val LEFT_CONTROLS = controls()
	val RIGHT_CONTROLS = controls()
	val TOP_CONTROLS = controls()
	val BOTTOM_CONTROLS = controls()
	val FRONT_CONTROLS = controls()
	val BACK_CONTROLS = controls()

	val CONTROLS = immutableMap {
		put(RelativeSide.BOTTOM, BOTTOM_CONTROLS)
		put(RelativeSide.TOP, TOP_CONTROLS)
		put(RelativeSide.LEFT, LEFT_CONTROLS)
		put(RelativeSide.RIGHT, RIGHT_CONTROLS)
		put(RelativeSide.FRONT, FRONT_CONTROLS)
		put(RelativeSide.BACK, BACK_CONTROLS)
	}

	val BATTERY_ONLY = controlsGrid.next()
	val ITEMS_CONFIGURATION = controlsGrid.next()
	val ENERGY_CONFIGURATION = controlsGrid.next()
	val FLUID_CONFIGURATION = controlsGrid.next()

	val LEFT_CONTROLS_ITEMS = controls2(LEFT_CONTROLS)
	val RIGHT_CONTROLS_ITEMS = controls2(RIGHT_CONTROLS)
	val TOP_CONTROLS_ITEMS = controls2(TOP_CONTROLS)
	val BOTTOM_CONTROLS_ITEMS = controls2(BOTTOM_CONTROLS)
	val FRONT_CONTROLS_ITEMS = controls2(FRONT_CONTROLS)
	val BACK_CONTROLS_ITEMS = controls2(BACK_CONTROLS)

	val ITEMS_CONTROLS = immutableMap {
		put(RelativeSide.BOTTOM, BOTTOM_CONTROLS_ITEMS)
		put(RelativeSide.TOP, TOP_CONTROLS_ITEMS)
		put(RelativeSide.LEFT, LEFT_CONTROLS_ITEMS)
		put(RelativeSide.RIGHT, RIGHT_CONTROLS_ITEMS)
		put(RelativeSide.FRONT, FRONT_CONTROLS_ITEMS)
		put(RelativeSide.BACK, BACK_CONTROLS_ITEMS)
	}
}
