package ru.dbotthepony.mc.otm.data.loot

import com.google.common.collect.ImmutableList
import com.google.gson.*
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.Dynamic
import com.mojang.serialization.JsonOps
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.storage.loot.Deserializers
import net.minecraft.world.level.storage.loot.LootContext
import net.minecraft.world.level.storage.loot.LootPool
import net.minecraft.world.level.storage.loot.predicates.LootItemCondition
import net.minecraftforge.common.ForgeHooks
import net.minecraftforge.common.loot.IGlobalLootModifier
import net.minecraftforge.common.loot.LootModifier
import java.util.Arrays
import java.util.Deque
import java.util.stream.Stream

class LootPoolAppender(conditions: Array<out LootItemCondition>, pools: Stream<LootPool>) : LootModifier(conditions) {
	private val pools = pools.collect(ImmutableList.toImmutableList())

	constructor(conditions: Array<out LootItemCondition>, pools: Collection<LootPool>) : this(conditions, pools.stream())
	constructor(conditions: Array<out LootItemCondition>, vararg pools: LootPool) : this(conditions, Arrays.stream(pools))

	override fun doApply(generatedLoot: ObjectArrayList<ItemStack>, context: LootContext): ObjectArrayList<ItemStack> {
		for (pool in pools) {
			pool.addRandomItems(generatedLoot::add, context)
		}

		return generatedLoot
	}

	override fun codec(): Codec<out IGlobalLootModifier> {
		return CODEC
	}

	// TODO: remove reflection once Forge implement a way to provide lootpool deserialization context in non-reflective way
	companion object {
		private val lootPoolCodec: Codec<List<LootPool>>

		init {
			val serializer = Deserializers.createLootTableSerializer().create()

			lootPoolCodec = Codec.list(Codec.PASSTHROUGH.flatXmap({
				try {
					DataResult.success(serializer.fromJson(it.convert(JsonOps.INSTANCE).value, LootPool::class.java))
				} catch(err: JsonSyntaxException) {
					DataResult.error { err.message }
				}
			}, {
				try {
					DataResult.success(Dynamic(JsonOps.INSTANCE, serializer.toJsonTree(it)))
				} catch(err: JsonSyntaxException) {
					DataResult.error { err.message }
				}
			}))
		}

		val CODEC: Codec<LootPoolAppender> =
			RecordCodecBuilder.create {
				codecStart(it).and(
					lootPoolCodec.fieldOf("pools").forGetter(LootPoolAppender::pools)
				).apply(it, ::LootPoolAppender)
			}
	}
}
